package classinfo;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;


/**
 * Representation of a class file.
 *
 * @author SHUDOH Kazuyuki
 */
public
class Classfile {
  static final int magicNumber = 0xcafebabe;	// 4 bytes
  int minorVersion;	// 2 bytes
  int majorVersion;	// 2 bytes
  int CPCount;		// 2 bytes
  Vector constantPool = new Vector();
  int accessFlags;	// 2 bytes
  int thisClass;	// 2 bytes
  int superClass;	// 2 bytes
  int interfaceCount;	// 2 bytes
  int interfaces[];	// each entry is 2 bytes
  int fieldCount;	// 2 bytes
  Vector fieldTable = new Vector();
  int methodCount;	// 2 bytes
  Vector methodTable = new Vector();
  int attributesCount;	// 2 bytes
  Vector attributes = new Vector();

  /**
   * Reads a class file from in and constructs an instance of this class.
   */
  public Classfile(DataInputStream in)
	throws IOException, ClassFormatError {
    BufManager bufm = new BufManager(in);

try{
    bufm.advance(4);
    if (bufm.parseInt(0, 4) != magicNumber) {
      throw new ClassFormatError("Illegal magic number");
    }

    // minor, major number, constant pool count
    bufm.clearBuffer();
    bufm.advance(6);

    minorVersion = bufm.parseInt(0, 2);
    majorVersion = bufm.parseInt(2, 2);
    CPCount = bufm.parseInt(4, 2);

    // constant pool
    CPParse(bufm, CPCount, constantPool);

    // access flag, this class, super class, interfaces count
    bufm.clearBuffer();
    bufm.advance(8);

    accessFlags = bufm.parseInt(0, 2);
    thisClass = bufm.parseInt(2, 2);
    superClass = bufm.parseInt(4, 2);
    interfaceCount = bufm.parseInt(6, 2);
    interfaces = new int[interfaceCount];

    // interfaces
    for (int i = 0; i < interfaceCount; i++) {
      bufm.clearBuffer();
      bufm.advance(2);
      interfaces[i] = bufm.parseInt(0, 2);
    }

    // fields count
    bufm.clearBuffer();
    bufm.advance(2);

    fieldCount = bufm.parseInt(0, 2);

    // fields
    for (int i = 0; i < fieldCount; i++) {
      fieldTable.addElement(new Field(bufm, constantPool));
    }

    // methods count
    bufm.clearBuffer();
    bufm.advance(2);

    methodCount = bufm.parseInt(0, 2);

    // methods
    for (int i = 0; i < methodCount; i++) {
      methodTable.addElement(new Method(bufm, constantPool));
    }

    // attributes count
    bufm.clearBuffer();
    bufm.advance(2);

    attributesCount = bufm.parseInt(0, 2);

    // attributes
    for (int i = 0; i < attributesCount; i++) {
      bufm.clearBuffer();
      bufm.advance(6);
      attributes.addElement(new Attribute(bufm.parseInt(0, 2),
					bufm.parseInt(2, 4),
					bufm, constantPool));
    }
}
catch (Exception e) {
e.printStackTrace();
}
  }


  /**
   * Writes this class as class file into a data output stream, out.
   *
   * @param out a data output stream
   */
  public void write(DataOutputStream out)
	throws java.io.IOException {
    out.writeInt(magicNumber);

    out.writeShort((short)minorVersion);
    out.writeShort((short)majorVersion);
    out.writeShort((short)CPCount);

    for (int i = 1; i < CPCount; i++) {
      ((ConstantInPool)constantPool.elementAt(i)).write(out);
    }

    out.writeShort((short)accessFlags);
    out.writeShort((short)thisClass);
    out.writeShort((short)superClass);
    out.writeShort((short)interfaceCount);

    for (int i =0; i < interfaceCount; i++) {
      out.writeShort((short)interfaces[i]);
    }

    out.writeShort((short)fieldCount);
    for (Enumeration e = fieldTable.elements(); e.hasMoreElements(); ) {
      ((Field)e.nextElement()).write(out);
    }

    out.writeShort((short)methodCount); 
    for (Enumeration e = methodTable.elements(); e.hasMoreElements(); ) {
      ((Method)e.nextElement()).write(out);
    }

    out.writeShort((short)attributesCount);
    for (Enumeration e = attributes.elements(); e.hasMoreElements(); ) {
      ((Attribute)e.nextElement()).write(out);
    }
  }


  /**
   * Write this class as text into a print stream, out.
   *
   * @param out a print stream
   */
  public void list(PrintStream out) {
    out.println("Version: " + majorVersion + "." + minorVersion);

    out.println("Constant pool count: " + CPCount);
    out.println("Constant pool:");
    {
      Enumeration e = constantPool.elements();
      e.nextElement();	// skip constantPool.elementAt(0)

      int count = 1;
      while (e.hasMoreElements()) {
	out.print("#" + count++ + " ");
	out.print(((ConstantInPool)e.nextElement()).toString());
      }
    }

    out.println(
	"Access flag: 0x" + BufManager.intToHexString(accessFlags,4) +
	" (" + AccessFlags.toString((short)accessFlags) + ")\n");
    out.println("This class: " + thisClass);
    out.println("Super class: " + superClass);

    out.println("Interface count: " + interfaceCount);
    if (interfaceCount > 0) {
      out.println("Interfaces:");
      for (int i = 0; i < interfaceCount; i++) {
	out.println("  " + interfaces[i]);
      }
    }

    out.println("Field count: " + fieldCount);
    for (Enumeration e = fieldTable.elements(); e.hasMoreElements(); ) {
      out.print(((Field)e.nextElement()).toString());
    }

    out.println("Method count: " + methodCount);
    for (Enumeration e = methodTable.elements(); e.hasMoreElements(); ) {
      out.print(((Method)e.nextElement()).toString());
    }

    out.println("Attributes count: " + attributesCount);
    for (Enumeration e = attributes.elements(); e.hasMoreElements(); ) {
      out.print(((Attribute)e.nextElement()).toString());
    }

    out.flush();
  }


  /**
   * get Constant Pool count.
   */
  public int constantPoolCount() { return CPCount; }
  /**
   * get Constant Pool element.
   */
  public ConstantInPool constant(int index) {
    return (ConstantInPool)this.constantPool.elementAt(index);
  }

  /**
   * get Constant Pool element represents this class.
   */
  public ConstantInPool thisClass() { return constant(this.thisClass); }

  /**
   * get Constant Pool element represents superclass.
   */
  public ConstantInPool superClass() { return constant(this.superClass); }

  /**
   * get interface count of this class.
   */
  public int interfaceCount() { return this.interfaceCount; }
  /**
   * get a Constant Pool element rpresents an interface.
   */
  public ConstantInPool getInterface(int index) {
    return constant(this.interfaces[index]);
  }

  /**
   * get field count of this class.
   */
  public int fieldCount() { return this.fieldCount; }
  /**
   * get a field elemnt.
   */
  public Field field(int index) {
    return (Field)this.fieldTable.elementAt(index);
  }

  /**
   * get method count of this class.
   */
  public int methodCount() { return this.methodCount; }
  /**
   * get a method element.
   */
  public Method method(int index) {
    return (Method)this.methodTable.elementAt(index);
  }

  /**
   * get attributes count of this class.
   */
  public int attributeCount() { return this.attributesCount; }
  /**
   * get an attribute element.
   */
  public Attribute attribute(int index) {
    return (Attribute)this.attributes.elementAt(index);
  }


  /**
   * Replaces the constant at the specified index with the specified constant.
   */
  public void setConstantAt(ConstantInPool constant, int index) {
    this.constantPool.setElementAt(constant, index);
  }


  /**
   * Inserts the specified constant at the specified index.
   */
  public void insertConstantAt(ConstantInPool constant, int index) {
    this.constantPool.insertElementAt(constant, index);
    this.CPCount++;

    // adjust indices
    Enumeration e = constantPool.elements();
    e.nextElement();	// skip constantPool.elementAt(0)
    while (e.hasMoreElements()) {
      ConstantInPool constElem = (ConstantInPool)e.nextElement();
      ConstantPoolReference[] refs = constElem.constantPoolRefs();
      for (int i = 0; i < refs.length; i++) {
	int oldIndex = refs[i].index();
	if (oldIndex >= index) {
	  refs[i].index(oldIndex + 1);
	}
      }
    }
  }


  /**
   * Removes the constant at the specified index.
   *
   * @return an array of indices of constants
   *         which refer to the removed constant.
   */
  public int[] removeConstantAt(int index) {
    int[] indices = new int[this.CPCount];
    int indicesCount = 0;

    this.constantPool.removeElementAt(index);
    this.CPCount--;

    // adjust indices
    for (int i = 1; i < this.CPCount; i++) {
      ConstantInPool constElem = (ConstantInPool)constantPool.elementAt(i);
      ConstantPoolReference[] refs = constElem.constantPoolRefs();
      for (int j = 0; j < refs.length; j++) {
	int ind = refs[j].index();
	if (ind > index) {
	  refs[j].index(ind - 1);
	}
	else if (ind == index) {
	  indices[indicesCount++] = i;
	}
      }
    }


    int[] ret = new int[indicesCount];
    System.arraycopy(indices, 0, ret, 0, indicesCount);

    return ret;
  }


  private void CPParse(BufManager bufm, int CPCount, Vector CP)
	throws IOException {
    // dummy constant in pool[0]
    CP.addElement(new ConstantInPool(0, "All exceptions"));

    for (int count = 1; count < CPCount; count++) {
      ConstantInPool cp = new ConstantInPool(bufm, CP);
      CP.addElement(cp);

      if (cp.is64bit()) {
	CP.addElement(new ConstantInPool());	// null element
	count++;
      }
    }
  }


  private static void usage(String cmd) {
    System.out.println("usage:");
    System.out.println("\t" + cmd + " classfile [classfile ...]");
  }


  /**
   * Reads the classfile given as arguments
   * and writes as text to standard output.
   */
  public static void main(String argv[]) throws IOException {
    final String COMMAND = "classinfo";
    Classfile cinfo = null;

    if (argv.length <= 0) {
      usage(COMMAND);
      System.exit(1);
    }

    for (int i = 0; i < argv.length; i++) {
      try {
	DataInputStream in = new DataInputStream(new FileInputStream(argv[i]));
	cinfo = new Classfile(in);
//cinfo.setConstantAt(new ConstantInPool("bar"), 43);
	cinfo.list(System.out);

/*
	FileOutputStream fout = new FileOutputStream("classout" + i);
	DataOutputStream out = new DataOutputStream(fout);
	cinfo.write(out);
	out.flush();
	fout.close();
*/
      }
      catch (FileNotFoundException e) {
	System.err.println(e);
      }
    }
  }
}
